"""
reports.py - contains functions to work with reports in VMDR.
"""

from typing import Literal, Union
from os.path import join, exists
from os import mkdir
from io import StringIO

from requests.models import Response
from pandas import DataFrame, read_csv

from .data_classes import VMDRReport, ReportTemplate, VMDRScheduledReport
from ..auth import BasicAuth
from ..base import call_api
from ..base import xml_parser
from ..base.base_list import BaseList
from ..exceptions.Exceptions import *


def manage_scheduled_reports(
    auth: BasicAuth,
    action: Literal["list", "launch_now"],
    **kwargs,
) -> Response:
    """
    Backend function to manage scheduled reports in Qualys VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        action: Literal["list", "launch"] - The action to take.

    Returns:
        Response - The response from the API.

    :Kwargs:
        Look at the specific functions for the kwargs they accept: get_scheduled_report_list, launch_scheduled_report, etc.
    """

    # Set specific kwargs
    kwargs["action"] = action
    kwargs["echo_request"] = False

    headers = {"X-Requested-With": "qualysdk SDK"}

    match action:
        case "list":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="get_scheduled_report_list",
                params=kwargs,
                headers=headers,
            )

        case "launch_now":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="launch_scheduled_report",
                payload=kwargs,
                headers=headers,
            )

        case _:
            raise NotImplementedError(f"Action {action} is not valid.")


def manage_reports(
    auth: BasicAuth,
    action: Literal["list", "launch", "cancel", "fetch", "delete"],
    **kwargs,
) -> Response:
    """
    Backend function to manage reports in Qualys VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        action: Literal["list", "launch", "cancel", "fetch", "delete"] - The action to take.

    Returns:
        Response - The response from the API.

    :Kwargs:
        Look at the specific functions for the kwargs they accept: get_report_list, launch_report, etc.
    """

    # Set specific kwargs
    kwargs["action"] = action
    kwargs["echo_request"] = False

    headers = {"X-Requested-With": "qualysdk SDK"}

    match action:
        case "list":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="get_report_list",
                params=kwargs,
                headers=headers,
            )

        case "launch":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="launch_report",
                payload=kwargs,
                headers=headers,
            )

        case "cancel":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="cancel_report",
                payload=kwargs,
                headers=headers,
            )

        case "fetch":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="fetch_report",
                params=kwargs,
                headers=headers,
            )

        case "delete":
            return call_api(
                auth=auth,
                module="vmdr",
                endpoint="delete_report",
                payload=kwargs,
                headers=headers,
            )

        case _:
            raise NotImplementedError(f"Action {action} is not valid.")


def get_report_list(auth: BasicAuth, **kwargs) -> BaseList[VMDRReport]:
    """
    Get a list of reports in VMDR, according to kwargs.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.

    :Kwargs:
        id: Optional[Union[int,str]] - A specific report ID to get.
        state: Optional[str] - Filter output to reports in a specific state.
        user_login: Optional[str] - Filter output to reports launched by a specific user.
        expires_before_datetime: Optional[str] - Filter output to reports that will expire before this datetime. Formatted like: YYYY-MM-DD[THH:MM:SSZ]
        client_id: Optional[Union[int,str]] - Filter output to reports for a specific client ID. ONLY VALID FOR CONSULTANT SUBSCRIPTIONS!
        client_name: Optional[str] - Filter output to reports for a specific client name. ONLY VALID FOR CONSULTANT SUBSCRIPTIONS!

    Returns:
        BaseList[VMDRReport] - A list of VMDRReport objects.
    """

    response = manage_reports(auth, action="list", **kwargs)

    data = xml_parser(response.text)

    reports = data["REPORT_LIST_OUTPUT"]["RESPONSE"]["REPORT_LIST"]["REPORT"]

    bl = BaseList()

    # Check if there are multiple reports or just one
    if isinstance(reports, dict):
        reports = [reports]

    for report in reports:
        bl.append(VMDRReport.from_dict(report))

    return bl


def get_template_list(auth: BasicAuth) -> BaseList[ReportTemplate]:
    """
    Get the list of report templates in your subscription.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.

    Returns:
        BaseList[ReportTemplate] - A list of ReportTemplate objects.
    """

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="get_template_list",
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    data = xml_parser(response.text)

    if "SIMPLE_RETURN" in data:
        raise QualysAPIError(data["SIMPLE_RETURN"]["RESPONSE"]["TEXT"])

    bl = BaseList()

    templates = data["REPORT_TEMPLATE_LIST"]["REPORT_TEMPLATE"]

    # Check if there are multiple templates or just one
    if isinstance(templates, dict):
        templates = [templates]

    for template in templates:
        bl.append(ReportTemplate.from_dict(template))

    return bl


def launch_report(auth: BasicAuth, template_id: str, **kwargs) -> int:
    """
    Generate a new report in VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        template_id: Union[int, str] - The ID of the template to use for the report.

    :Kwargs:

        Parameter| Possible Values |Description|Required|
        |--|--|--|--|
        |auth|qualysdk.auth.BasicAuth|The authentication object.|✅|
        |template_id|Union[int, str] |The template that the report will follow. Use get_report_template_list() To select one.|✅|
        |report_title|str|The name to give to the report. ⚠️ IF YOU REQUEST A PCI COMPLIANCE REPORT, THE TITLE IS AUTO-GENERATED BY QUALYS!|❌|
        |output_format| FOR MAP REPORT: <br> pdf, html, mht, xml, csv<br>FOR SCAN REPORT:<br>pdf, html mht, xml, csv, docx<br>FOR REMEDIATION REPORT:<br>pdf, html, mht, csv<br>FOR COMPLIANCE (NON-PCI) REPORT:<br>pdf, html, mht<br>FOR PCI COMPLIANCE REPORT:<br>pdf, html<br>FOR PATCH REPORT:<br>pdf, online, xml, csv<br>FOR COMPLIANCY POLICY REPORT:<br>pdf, html, mht, xml, csv|The format that the report will be generated in.|❌|
        |hide_header|True/False| ⚠️ SDK auto-sets this to True!|❌|
        |pdf_password|str|If output_format==pdf, file will be encrypted with this password. Note that this is required for recipient_group/recipient_group_id. <br>⚠️ REQUREMENTS:<br>1.8<=N<=32 characters<br>2. Must contain alpha and numeric characters<br>3.Cannot match your Qualys account's password<br>4.Must follow any other password restrictions in Users->Setup->Security|❌|
        |recipient_group|str: "groupOne,GroupTwo"|A comma-separated string of group that the PDF will be shared with. ⚠️ CANNOT BE IN THE SAME REQUEST WITH recipient_group_id|❌|
        |recipient_group_id|str|A comma-separated string of group IDs to share the PDF with. ⚠️ CANNOT BE IN THE SAME REQUEST WITH recipient_group| ❌|
        |report_type|Literal["Map", "Scan", "Patch", "Remediation", "Compliance", "Policy"]|Shape the report to a specific type.|❌|
        |domain|str| Target domain for the report.|⚠️ REQUIRED FOR MAP REPORT|
        |ip_restriction|Comma-separated string of IP addresses to include in a map report.|⚠️ REQUIRED FOR MAP REPORT WHEN domain=='None'|
        |report_refs|str|Comma-separated string of reference IDs.|⚠️ REQUIRED FOR MAP REPORT, MANUAL SCAN REPORT, PCI COMPLIANCE REPORT|
        |asset_group_ids|str|Override asset group IDs defined in the report template with these IDs.|❌|
        |ips_network_id|Union[int, str]|Restrict the report to specific network IDs. ⚠️ MUST BE ENABLED IN THE QUALYS SUBSCRIPTION|❌|
        |ips|str|Comma-separated string of IP addresses to include, overwriting the report template.|❌|
        |assignee_type|Literal["User", "All"]|Specify if tickets assigned to the requesting user, or all tickets will be included in the report. Defaults to "User".|❌|
        |policy_id|Union[int, str]|The specific policy to run the report on.|❌|
        |host_id|str|In policy report output, show results for a single host. |⚠️ REQUIRED WHEN instance_string IS SPECIFIED.|
        |instance_string|str|Specifies a single instance on a host machine.|⚠️ REQUIRED WHEN host_id IS SPECIFIED.|

    Returns:
        int - The ID of the report.
    """

    # Set specific kwargs
    kwargs["template_id"] = template_id
    kwargs["hide_header"] = True

    response = manage_reports(auth, action="launch", **kwargs)

    data = xml_parser(response.text)

    try:
        return int(data["SIMPLE_RETURN"]["RESPONSE"]["ITEM_LIST"]["ITEM"]["VALUE"])
    except KeyError:
        raise QualysAPIError(data["SIMPLE_RETURN"]["RESPONSE"]["TEXT"])


def cancel_report(auth: BasicAuth, id: Union[int, str]) -> str:
    """
    Cancel a report in VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        id: Union[int, str] - The ID of the report to cancel.

    Returns:
        str - The response from the API.
    """

    response = manage_reports(auth, action="cancel", id=id)

    data = xml_parser(response.text)

    return data["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]


def fetch_report(
    auth: BasicAuth, id: Union[int, str], write_out: bool = False
) -> Union[DataFrame, None]:
    """
    Fetch a report in VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        id: Union[int, str] - The ID of the report to fetch.
        write_out: Optional[bool] - If True, write the report to ./reports/<file>. Default is False.

    Returns:
        Union[DataFrame, None] - The report as a DataFrame if it is a CSV or XML report. Otherwise, None.
    """
    return_data = False

    response = manage_reports(auth, action="fetch", id=id)

    file_format = response.headers["Content-Type"].split("/")[-1]
    if ";" in file_format:
        file_format = file_format.split(";")[0]

    # Special case for octet-stream/word documents
    if "octet" in file_format:
        file_format = "docx"

    # Check that there is a reports directory in the directory __file__ is in
    curr_dir = join(__file__, "..")

    # Match the file format and load into the appropriate object
    match file_format:
        case "csv":
            print("Detected CSV format. Returning DataFrame.")
            # Check for a summarization header. If it exists, skip it.
            # We can delete everything before "\r\n\r\n\r\n"
            content = response.text
            count = content.count("\r\n\r\n\r\n")
            if count > 0:
                # Special case for reports with a small table before the actual CSV data starts:
                split_content = content.split("\r\n\r\n\r\n")
                if (
                    '"Severity","Total","Confirmed","Potential","Information Gathered"\r\n'
                    in split_content[~0]
                ):
                    # Split again to get the actual data, this time with 2 \r\n's:
                    content = split_content[~0].split("\r\n\r\n", 1)[
                        ~0
                    ]  # maxsplit lets us ignore newlines except for the table separator
                else:
                    content = split_content[
                        ~0
                    ]  # get the last element, which is the actual data...
            data = read_csv(StringIO(content))
            return_data = True

        case "xml":
            print("Detected XML format. Returning DataFrame.")
            data = DataFrame.from_dict(xml_parser(response.text))
            return_data = True

        case _:
            print(
                f"Detected {file_format} format. Writing to {join(curr_dir, 'output', f'{id}.{file_format}')}"
            )
            write_out = True

    if write_out:
        if not exists(join(curr_dir, "output")):
            mkdir(join(curr_dir, "output"))
            print(f"Created output directory at {join(curr_dir, 'output')}")

        with open(join(curr_dir, "output", f"{id}.{file_format}"), "wb") as f:
            f.write(response.content)
            print(f"Wrote report to {join(curr_dir, 'output', f'{id}.{file_format}')}")

    if return_data:
        return data


def delete_report(auth: BasicAuth, id: Union[int, str]) -> str:
    """
    Delete a report in VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        id: Union[int, str] - The ID of the report to delete.

    Returns:
        str - The response from the API.
    """

    response = manage_reports(auth, action="delete", id=id)

    data = xml_parser(response.text)

    return data["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]


def get_scheduled_report_list(auth: BasicAuth, **kwargs) -> BaseList[VMDRReport]:
    """
    Get a list of scheduled reports in VMDR, according to kwargs.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.

    :Kwargs:
        id: Optional[Union[int,str]] - A specific report ID to get.
        is_active: Optional[bool] - Filter output to active or inactive reports. True for active, False for inactive.

    Returns:
        BaseList[VMDRReport] - A list of VMDRReport objects.
    """

    response = manage_scheduled_reports(auth, action="list", **kwargs)

    data = xml_parser(response.text)

    bl = BaseList()

    try:
        reports = data["SCHEDULE_REPORT_LIST_OUTPUT"]["RESPONSE"][
            "SCHEDULE_REPORT_LIST"
        ]["REPORT"]
        # Check if there are multiple reports or just one
        if isinstance(reports, dict):
            reports = [reports]

        for report in reports:
            bl.append(VMDRScheduledReport.from_dict(report))
    except KeyError:
        print("No reports found.")

    return bl


def launch_scheduled_report(auth: BasicAuth, id: str) -> str:
    """
    Launch a scheduled report in VMDR.

    Parameters:
        auth: Required[BasicAuth] - The BasicAuth object.
        id: str - The ID of the scheduled report to launch.

    Returns:
        str - The response from the API.
    """

    response = manage_scheduled_reports(auth, action="launch_now", id=id)

    data = xml_parser(response.text)

    return data["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]



"""
Cloud Agent APIs let you pull data on and modify Cloud Agents in your account.

https://cdn2.qualys.com/docs/qualys-ca-api-user-guide.pdf
"""

from .purge import purge_agent, bulk_purge_agent
from .calls import list_agents, launch_ods, bulk_launch_ods



"""
Contains the Agent data class for the Cloud Agent module.
"""

from typing import List, Union, Literal
from datetime import datetime
from dataclasses import dataclass, asdict, field
from ipaddress import ip_address

from ...base.base_list import BaseList
from ..data_classes.CloudAgentTag import CloudAgentTag


@dataclass
class CloudAgent:
    """
    Represents one cloud agent in a Qualys subscription.
    """

    id: int = field(default=None)
    name: str = field(default=None)
    created: datetime = field(default=None)
    modified: datetime = field(default=None)
    type: str = field(default=None)
    tags: BaseList[CloudAgentTag] = field(default=None)
    sourceInfo: list = field(default=None)
    qwebHostId: int = field(default=None)
    lastComplianceScan: datetime = field(default=None)
    lastSystemBoot: datetime = field(default=None)
    lastLoggedOnUser: str = field(default=None)
    dnsHostName: str = field(default=None)
    agentInfo: None = field(default=None)
    # agentInfo is parsed out into below fields:
    agentInfo_agentVersion: str = field(default=None)
    agentInfo_agentId: str = field(default=None)
    agentInfo_status: str = field(default=None)
    agentInfo_lastCheckedIn: datetime = field(default=None)
    agentInfo_connectedFrom: ip_address = field(default=None)
    agentInfo_location: str = field(default=None)
    agentInfo_locationGeoLatitude: float = field(default=None)
    agentInfo_locationGeoLongitude: float = field(default=None)
    agentInfo_chripStatus: str = field(default=None)
    agentInfo_platform: str = field(default=None)
    agentInfo_activatedModule: str = field(default=None)
    agentInfo_manifestVersion: str = field(default=None)
    agentInfo_agentConfiguration: None = field(default=None)
    # agentConfiguration is parsed out into below fields,
    # and is a sub-dictionary of agentInfo:
    agentInfo_agentConfiguration_id: int = field(default=None)
    agentInfo_agentConfiguration_name: str = field(default=None)
    # END agentConfiguration fields
    agentInfo_activationKey: None = field(default=None)
    # agentInfo_activationKey is a sub-dictionary of agentInfo
    # and is parsed out into below fields:
    agentInfo_activationKey_activationId: str = field(default=None)
    agentInfo_activationKey_title: str = field(default=None)
    # END agentInfo_activationKey fields
    # END agentInfo fields
    netbiosName: str = field(default=None)
    criticalityScore: int = field(default=None)
    lastVulnScan: datetime = field(default=None)
    vulnsUpdated: datetime = field(default=None)
    informationGatheredUpdated: datetime = field(default=None)
    domain: str = field(default=None)
    fqdn: str = field(default=None)
    os: str = field(default=None)
    networkGuid: str = field(default=None)
    address: ip_address = field(default=None)
    trackingMethod: Literal[
        "NONE", "IP", "DNSNAME", "NETBIOS", "INSTANCE_ID", "QAGENT"
    ] = field(default=None)
    manufacturer: str = field(default=None)
    model: str = field(default=None)
    totalMemory: int = field(default=None)
    timezone: str = field(default=None)
    biosDescription: str = field(default=None)
    openPort: BaseList[str] = field(default=None)
    software: BaseList[str] = field(default=None)
    vuln: BaseList[str] = field(default=None)
    processor: str = field(default=None)
    volume: BaseList[str] = field(default=None)
    account: BaseList[str] = field(default=None)
    networkInterface: BaseList[str] = field(default=None)
    isDockerHost: bool = field(default=None)
    dockerInfo: str = field(default=None)
    cloudProvider: str = field(default=None)

    def __post_init__(self):
        DATE_FIELDS = [
            "created",
            "modified",
            "lastComplianceScan",
            "lastSystemBoot",
            "agentInfo_lastCheckedIn",
            "lastVulnScan",
            "vulnsUpdated",
            "informationGatheredUpdated",
        ]
        INT_FIELDS = [
            "id",
            "qwebHostId",
            "agentInfo_agentConfiguration_id",
            "criticalityScore",
            "totalMemory",
        ]
        TO_NONE_FIELDS = [
            "agentInfo",
            "agentInfo_agentConfiguration",
            "agentInfo_activationKey",
        ]
        FLOAT_FIELDS = [
            "agentInfo_locationGeoLatitude",
            "agentInfo_locationGeoLongitude",
        ]
        IP_ADDRESS_FIELDS = ["agentInfo_connectedFrom"]

        for field in DATE_FIELDS:
            if getattr(self, field):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        for field in INT_FIELDS:
            if getattr(self, field):
                setattr(self, field, int(getattr(self, field)))

        for field in FLOAT_FIELDS:
            if getattr(self, field):
                setattr(self, field, float(getattr(self, field)))

        for field in IP_ADDRESS_FIELDS:
            if getattr(self, field):
                setattr(self, field, ip_address(getattr(self, field)))

        # Before setting none fields, let's build the BaseList objects
        if self.tags:
            tags = BaseList()
            data = self.tags.get("list")
            if data:
                data = data.get("TagSimple")
                if data and not isinstance(data, list):
                    data = [data]
                for tag in data:
                    tags.append(CloudAgentTag(**tag))

            if len(tags) > 0:
                setattr(self, "tags", tags)
            else:
                setattr(self, "tags", None)

        if self.openPort:
            open_ports = BaseList()
            data = self.openPort.get("list")
            if data:
                data = data.get("HostAssetOpenPort")
                if data and not isinstance(data, list):
                    data = [data]
                for port in data:
                    open_ports.append(
                        f"{port.get('port')}-{port.get('protocol')} ({port.get('serviceName')})"
                    )

            if len(open_ports) > 0:
                setattr(self, "openPort", open_ports)
            else:
                setattr(self, "openPort", None)

        if self.software:
            software = BaseList()
            data = self.software.get("list")
            if data:
                data = data.get("HostAssetSoftware")
                if data and not isinstance(data, list):
                    data = [data]
                for software_item in data:
                    software.append(
                        f"{software_item.get('name')} ({software_item.get('version')})"
                    )

            if len(software) > 0:
                setattr(self, "software", software)
            else:
                setattr(self, "software", None)

        if self.vuln:
            vulns = BaseList()
            data = self.vuln.get("list")
            if data:
                data = data.get("HostAssetVuln")
                if data and not isinstance(data, list):
                    data = [data]
                for vuln in data:
                    vulns.append(
                        f"{vuln.get('qid')} ({vuln.get('hostInstanveVulnId')}) First found: {vuln.get('firstFound')} Last found: {vuln.get('lastFound')}"
                    )

            if len(vulns) > 0:
                setattr(self, "vuln", vulns)
            else:
                setattr(self, "vuln", None)

        if self.processor:
            temp = []
            data = self.processor.get("list")
            if data:
                data = data.get("HostAssetProcessor")
                if data and not isinstance(data, list):
                    data = [data]
                for processor_item in data:
                    temp.append(f"{processor_item.get('name')}")

            if len(temp) > 0:
                setattr(self, "processor", temp[0])
            else:
                setattr(self, "processor", None)

        if self.volume:
            volumes = BaseList()
            data = self.volume.get("list")
            if data:
                data = data.get("HostAssetVolume")
                if data and not isinstance(data, list):
                    data = [data]
                for volume in data:
                    # Calculate the % free space on the volume. Account for 0 division!
                    try:
                        percent_free = (
                            int(volume.get("free")) / int(volume.get("size"))
                        ) * 100
                    except ZeroDivisionError:
                        percent_free = 0
                    volumes.append(f"{volume.get('name')} ({percent_free:.2f}% free)")

            if len(volumes) > 0:
                setattr(self, "volume", volumes)
            else:
                setattr(self, "volume", None)

        if self.account:
            accounts = BaseList()
            data = self.account.get("list")
            if data:
                data = data.get("HostAssetAccount")
                if data and not isinstance(data, list):
                    data = [data]
                for account in data:
                    accounts.append(account.get("username"))

            if len(accounts) > 0:
                setattr(self, "account", accounts)
            else:
                setattr(self, "account", None)

        if self.networkInterface:
            interfaces = BaseList()
            data = self.networkInterface.get("list")
            if data:
                data = data.get("HostAssetInterface")
                if data and not isinstance(data, list):
                    data = [data]
                for interface in data:
                    interfaces.append(
                        f"{interface.get('interfaceName')} ({interface.get('macAddress')})"
                    )

            if len(interfaces) > 0:
                setattr(self, "networkInterface", interfaces)
            else:
                setattr(self, "networkInterface", None)

        for field in TO_NONE_FIELDS:
            if getattr(self, field) == {}:
                setattr(self, field, None)

        if "isDockerHost" in asdict(self).keys():
            setattr(self, "isDockerHost", self.isDockerHost == "true")

        if self.dockerInfo:
            # dockerInfo is a dictionary with various keys and values.
            # Since we never know how many keys there are, we will
            # have to dynamically build the string.
            s = ""
            for key, value in self.dockerInfo.items():
                s += f"{key}: {value}"
                # if this is not the last key, add a comma and space
                if key != list(self.dockerInfo.keys())[-1]:
                    s += ", "

            setattr(self, "dockerInfo", s)

    def to_dict(self):
        return asdict(self)

    def __dict__(self):
        return self.to_dict()

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    def __int__(self):
        return self.id

    def __str__(self) -> str:
        return self.name

    def created_after(self, date: datetime) -> bool:
        """
        Check if the agent was created after a certain date.

        Args:
            date (datetime): The date to check against.

        Returns:
            bool: True if the agent was created after the date, False otherwise.
        """
        return self.created > date

    def created_before(self, date: datetime) -> bool:
        """
        Check if the agent was created before a certain date.

        Args:
            date (datetime): The date to check against.

        Returns:
            bool: True if the agent was created before the date, False otherwise.
        """
        return self.created < date

    def modified_after(self, date: datetime) -> bool:
        """
        Check if the agent was modified after a certain date.

        Args:
            date (datetime): The date to check against.

        Returns:
            bool: True if the agent was modified after the date, False otherwise.
        """
        return self.modified > date

    def modified_before(self, date: datetime) -> bool:
        """
        Check if the agent was modified before a certain date.

        Args:
            date (datetime): The date to check against.

        Returns:
            bool: True if the agent was modified before the date, False otherwise.
        """
        return self.modified < date



"""
purge.py - Contains the user-facing functions for purging Cloud Agents singularly or in bulk.
"""

from ..base.call_api import call_api
from ..base.xml_parser import xml_parser
from ..auth.basic import BasicAuth
from .prepare_criteria import prepare_criteria


def purge_agent(auth: BasicAuth, asset_id: str) -> str:
    """
    Purge a single Cloud Agent from your subscription.

    Args:
        auth (BasicAuth): The authentication object containing the user's credentials.
        asset_id (str): The **Asset ID** of the Cloud Agent to purge.

    Returns:
        str: The response from the API call.
    """

    xml_data = (
        '<?xml version="1.0" encoding="UTF-8" ?> <ServiceRequest></ServiceRequest>'
    )

    payload = {
        "placeholder": asset_id,  # For formatting the URL
        "_xml_data": xml_data,  # call_api will override body with this
    }

    response = call_api(
        auth=auth,
        module="cloud_agent",
        endpoint="purge_agent",
        payload=payload,
    )

    data = xml_parser(response.text)

    return data.get("ServiceResponse").get("responseCode")


def bulk_purge_agent(auth: BasicAuth, **kwargs) -> str:
    """
    Delete hosts in bulk based on asset ID, asset name,
    creation date, updated date, tagName, or agentUuid.

    NOTE: IT IS **HIGHLY** RECOMMENDED TO USE THE **asset_id** PARAMETER AS THE ONLY FILTER.

    Args:
        auth (BasicAuth): The authentication object containing the user's credentials.
        **kwargs: The filters to apply to the bulk purge.

    ## Kwargs:

        asset_id (list[str]): A list of asset IDs to purge. HIGHLY RECOMMENDED.
        name (list[str]): A list of asset names to purge.
        created (Union[str, datetime.datetime]): The creation date of the asset. Formatted like ```YYYY-MM-DD[THH:MM:SSZ] as a str.
        updated (Union[str, datetime.datetime]): The updated date of the asset. Formatted like ```YYYY-MM-DD[THH:MM:SSZ] as a str.
        created_operator (Literal['GREATER', 'LESSER']): The operator to apply to the created date.
        updated_operator (Literal['GREATER', 'LESSER']): The operator to apply to the updated date.
        tagName (list[str]): A list of tag names to purge assets under.
        agentUuid (list[str]): A list of agent UUIDs to purge.

    Returns:
        str: The response from the API call.
    """

    payload = {"_xml_data": prepare_criteria(**kwargs)}

    response = call_api(
        auth=auth,
        module="cloud_agent",
        endpoint="bulk_purge_agent",
        payload=payload,
    )

    data = xml_parser(response.text)

    return (
        data.get("ServiceResponse").get("responseCode")
        if response.status_code == 200
        else "ERROR: "
        + xml_parser(response.text)
        .get("ServiceResponse")
        .get("responseErrorDetails")
        .get("errorMessage")
    )



"""
get_host_list_detections.py - contains the get_host_list_detections function for the qualysdk package.

This endpoint is used to get a list of hosts and their QID detections. The function is multithreaded and uses the hld_backend function to pull the data.
"""

from typing import Union
from threading import Thread

from .base.helpers import create_id_queue, thread_worker, prepare_args
from ..base.base_list import BaseList
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *


def get_hld(
    auth: BasicAuth,
    chunk_size: int = 3000,
    threads: int = 5,
    page_count: Union[int, "all"] = "all",
    chunk_count: Union[int, "all"] = "all",
    **kwargs,
) -> BaseList:
    """
    get_hld - get a list of hosts and their QID detections using multiple threads.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        chunk_size (int): The size of each chunk. Defaults to 3000.
        threads (int): The number of threads to use. Defaults to 5.
        page_count (Union[int, "all"]): The number of pages to retrieve. Defaults to "all".
        chunk_count (Union[int, "all"]): The number of chunks to retrieve. Defaults to "all".
        **kwargs: Additional keyword arguments to pass to the API.

    Kwargs:

        action (Optional[str]) #The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        echo_request (Optional[bool]) #Whether to echo the request. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        show_asset_id (Optional[bool]) #Whether to show the asset IDs. Default is 'False'. ends up being passed to API as 0 or 1.
        include_vuln_type (Optional[Literal["confirmed", "potential"]]) #The type of vulnerability to include. If not specified, both types are included.

        DETECTION FILTERS:
        show_results (Optional[bool]) #Whether to show the results. Default is True. Ends up being passed to the API as 0 or 1. WARNING: this SDK overwrites any value you pass with '1'. It is just recognized as valid for the sake of completeness.
        arf_kernel_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns for Linux kernels. 0 = don't filter, 1 = exclude kernel vulns that are not exploitable, 2 = only include kernel related vulns that are not exploitable, 3 = only include exploitable kernel vulns, 4 = only include kernel vulns. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        arf_service_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns found on running or nonrunning ports/services. 0 = don't filter, 1 = exclude service related vulns that are exploitable, 2 = only include service vulns that are exploitable, 3 = only include service vulns that are not exploitable, 4 = only include service vulns. If specified, results are in a host's <AFFECT_RUNNING_SERVICE> tag.
        arf_config_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns that can be vulnerable based on host config. 0 = don't filter, 1 = exclude vulns that are exploitable due to host config, 2 = only include config related vulns that are exploitable, 3 = only include config related vulns that are not exploitable, 4 = only include config related vulns. If specified, results are in a host's <AFFECT_EXPLOITABLE_CONFIG> tag.
        active_kernels_only (Optional[Literal[0,1,2,3]]) #Specify vulns related to running or non-running kernels. 0 = don't filter, 1 = exclude non-running kernels, 2 = only include vulns on non-running kernels, 3 = only include vulns with running kernels. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        output_format (Optional[Literal["XML", "CSV"]]) #The format of the output. Default is 'XML'. WARNING: this SDK will overwrite any value you pass with 'XML'. It is just recognized as valid for the sake of completeness.
        supress_duplicated_data_from_csv (Optional[bool]) #Whether to suppress duplicated data from CSV. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        truncation_limit (Optional[int]) #The truncation limit for a page. Default is 1000.
        max_days_since_detection_updated (Optional[int]) #The maximum number of days since the detection was last updated. For detections that have never changed, the value is applied to the last detection date.
        detection_updated_since (Optional[str]) #The date and time since the detection was updated.
        detection_updated_before (Optional[str]) #The date and time before the detection was updated.
        detection_processed_before (Optional[str]) #The date and time before the detection was processed.
        detection_processed_after (Optional[str]) #The date and time after the detection was processed.
        detection_last_tested_since (Optional[str]) #The date and time since the detection was last tested.
        detection_last_tested_since_days (Optional[int]) #The number of days since the detection was last tested.
        detection_last_tested_before (Optional[str]) #The date and time before the detection was last tested.
        detection_last_tested_before_days (Optional[int]) #The number of days before the detection was last tested.
        include_ignored (Optional[bool]) #Whether to include ignored detections. Default is False. Ends up being passed to the API as 0 or 1.
        include_disabled (Optional[bool]) #Whether to include disabled detections. Default is False. Ends up being passed to the API as 0 or 1.

        HOST FILTERS:
        ids (Optional[str]) #A comma-separated string of host IDs to include.
        id_min (Optional[Union[int,str]]) #The minimum host ID to include.
        id_max (Optional[Union[int,str]]) #The maximum host ID to include.
        ips (Optional[str]) #The IP address of the host to include. Can be a comma-separated string, and also supports ranges with a hyphen: 10.0.0.0-10.0.0.255.
        ipv6 (Optional[str]) #The IPv6 address of the host to include. Can be a comma-separated string. Does not support ranges.
        ag_ids (Optional[str]) #Show only hosts belonging to the specified asset group IDs. Can be a comma-separated string, and also supports ranges with a hyphen: 1-5.
        ag_titles (Optional[str]) #Show only hosts belonging to the specified asset group titles. Can be a comma-separated string.
        network_ids (Optional[str]) #Show only hosts belonging to the specified network IDs. Can be a comma-separated string.
        network_names (Optional[str]) #displays the name of the network corresponding to the network ID.
        vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        no_vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        max_days_since_last_vm_scan (Optional[int]) #The maximum number of days since the last VM scan.
        vm_processed_before (Optional[str]) #The date and time before the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_processed_after (Optional[str]) #The date and time after the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_before (Optional[str]) #The date and time before the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_after (Optional[str]) #The date and time after the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_before (Optional[str]) #The date and time before the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_after (Optional[str]) #The date and time after the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        status (Optional[Literal["New", "Active", "Fixed", "Re-Opened"]]) #The status of the detection.
        compliance_enabled (Optional[bool]) #Whether compliance is enabled. Default is False. Ends up being passed to the API as 0 or 1.
        os_pattern (Optional[str]) #PCRE Regex to match operating systems.

        QID FILTERS:
        qids (Optional[str]) #A comma-separated string of QIDs to include.
        severities (Optional[str]) #A comma-separated string of severities to include. Can also be a hyphenated range, i.e. '2-4'.
        filter_superseded_qids (Optional[bool]) #Whether to filter superseded QIDs. Default is False. Ends up being passed to the API as 0 or 1.
        include_search_list_titles (Optional[str]) #A comma-separated string of search list titles to include.
        exclude_search_list_titles (Optional[str]) #A comma-separated string of search list titles to exclude.
        include_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to include.
        exclude_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to exclude.

        ASSET TAG FILTERS:
        use_tags (Optional[bool]) #Whether to use tags. Default is False. Ends up being passed to the API as 0 or 1.
        tag_set_by (Optional[Literal['id','name']]) #When filtering on tags, whether to filter by tag ID or tag name.
        tag_include_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_exclude_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_set_include (Optional[str]) #A comma-separated string of tag IDs or names to include.
        tag_set_exclude (Optional[str]) #A comma-separated string of tag IDs or names to exclude.
        show_tags (Optional[bool]) #Whether to show tags. Default is False. Ends up being passed to the API as 0 or 1.

        QDS FILTERS:
        show_qds (Optional[bool]) #Whether to show QDS. Default is False. Ends up being passed to the API as 0 or 1.
        qds_min (Optional[int]) #The minimum QDS to include.
        qds_max (Optional[int]) #The maximum QDS to include.
        show_qds_factors (Optional[bool]) #Whether to show QDS factors. Default is False. Ends up being passed to the API as 0 or 1.

        EC2/AZURE METADATA FILTERS:
        host_metadata (Optional[Literal['all,'ec2', 'azure']]) #The type of metadata to include. Default is 'all'.
        host_metadata_fields (Optional[str]) #A comma-separated string of metadata fields to include. Use carefully.
        show_cloud_tags (Optional[bool]) #Whether to show cloud tags. Default is False. Ends up being passed to the API as 0 or 1.
        cloud_tag_fields (Optional[str]) #A comma-separated string of cloud tag fields to include. Use carefully.

    Returns:
        BaseList: A list of VMDRHost objects, with their DETECTIONS attribute populated.
    """

    prepare_args(
        auth=auth,
        chunk_size=chunk_size,
        threads=threads,
        page_count=page_count,
        chunk_count=chunk_count,
        ids=kwargs.get("ids"),
    )

    id_queue = create_id_queue(auth, chunk_size=chunk_size, ids=kwargs.get("ids", None))
    print(f"Starting get_hld with {threads} {'threads.' if threads > 1 else 'thread.'}")

    threads_list = []

    responses = BaseList()

    for i in range(threads):
        thread = Thread(
            target=thread_worker,
            args=(
                auth,
                id_queue,
                responses,
                page_count,
                chunk_count,
                "get_hld",
                kwargs,
            ),
        )
        threads_list.append(thread)
        thread.start()

    for thread in threads_list:
        thread.join()

    print("All threads have completed. Returning responses.")
    return responses



"""
Exceptions.py - Custom Exceptions for package.
"""


class AuthenticationError(Exception):
    """
    Basic exception class for qualysdk when dealing with authentication.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class InvalidCredentialsError(AuthenticationError):
    """
    Exception for when credentials are invalid.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class InvalidTokenError(AuthenticationError):
    """
    Exception for when token is invalid.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class InvalidAuthTypeError(AuthenticationError):
    """
    Exception for when auth type is invalid.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class AuthTypeMismatchError(AuthenticationError):
    """
    Exception for when auth type is invalid.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class InvalidEndpointError(Exception):
    """
    Exception for when endpoint is invalid.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class QualysAPIError(Exception):
    """
    Basic exception class for qualysdk.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)



"""
qds.py - contains the QDS dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass(order=True)
class QDS:
    """
    QDS - represents a single QDS score.

    made with order=True to allow for sorting of QDS scores.
    """

    SEVERITY: str = field(
        metadata={"description": "The rating of the QDS score."}, compare=False
    )
    SCORE: int = field(
        metadata={"description": "The actual score itself as an int."},
        default=0,
    )

    def __post_init__(self):
        # make sure that the SEVERITY is a str:
        if not isinstance(self.SEVERITY, str):
            raise TypeError(f"QDS SEVERITY must be a str, not {type(self.SEVERITY)}")
        # and that text is an int:
        if not isinstance(self.SCORE, int):
            raise TypeError(f"QDS SCORE must be an int, not {type(self.SCORE)}")

    def __str__(self):
        return str(self.SCORE)

    def __int__(self):
        return self.SCORE

    def __contains__(self, item):
        return item in self.SEVERITY or item in self.SCORE

    def copy(self):
        return QDS(SEVERITY=self.SEVERITY, SCORE=self.SCORE)

    def is_severity(self, severity: str):
        return self.SEVERITY == severity

    def get_severity(self):
        return self.SEVERITY

    def get_score(self):
        return self.SCORE

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a QDS object from a dictionary.

        Params:
            data (dict): The dictionary containing the data for the QDS object.

        Returns:
            QDS: The QDS object created from the dictionary.
        """
        required_keys = {"SEVERITY", "SCORE"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )
        return cls(**data)



"""
SQL server integration for qualysdk
"""

from .base import db_connect
from .vmdr import (
    upload_vmdr_ags,
    upload_vmdr_kb,
    upload_vmdr_hosts,
    upload_vmdr_ips,
    upload_vmdr_hld,
    upload_vmdr_scanners,
    upload_vmdr_static_search_lists,
    upload_vmdr_dynamic_search_lists,
    upload_vmdr_users,
    upload_vmdr_scan_list,
    upload_vmdr_report_list,
    upload_vmdr_scheduled_report_list,
    upload_vmdr_template_list,
    upload_vmdr_kb_qvs,
    upload_vmdr_activity_log,
)

from .gav import upload_gav_hosts

from .cloud_agent import upload_cloud_agents



"""
base.py - contains the base functionality for the SQL module of qualysdk.
"""

from datetime import datetime
from dataclasses import dataclass
from typing import Literal

from sqlalchemy import types

try:
    import pyodbc
except ImportError as e:
    raise Warning(
        "Pyodbc cannot be imported. If you are on Mac or Unix, this usually can be resolved by installing unixODBC:\n\nMac: brew install unixodbc\nUbuntu: sudo apt install unixodbc"
    ) from e

from pandas import DataFrame
from sqlalchemy import create_engine, Connection, types


def db_connect(
    host: str,
    db: str,
    username: str = None,
    password: str = None,
    driver: str = "ODBC Driver 17 for SQL Server",
    trusted_cnxn: bool = False,
    db_type: Literal["mssql", "mysql", "postgresql", "sqlite", "oracle"] = "mssql",
) -> Connection:
    """

    Generate a sqlalchemy Connection object to a SQL database.

    Args:
        host (str): The hostname of the SQL server.
        db (str): The name of the database.
        username (str): The username to use to connect to the database.
        password (str): The password to use to connect to the database.
        driver (str): The ODBC driver to use to connect to the database. Defaults to 'ODBC Driver 17 for SQL Server'.
        trusted_cnxn (bool): If True, use trusted connection. If False, use username and password. Defaults to False.
        db_type (str): The type of database to connect to. Defaults to 'mssql'. Options are 'mssql', 'mysql', 'postgresql', 'sqlite', 'oracle'.

    Returns:
        Connection: The Connection object to the SQL database.
    """

    # Check if user AND password are provided, or trusted connection is used:
    if not (username and password) and not trusted_cnxn:
        raise ValueError(
            "You must provide a username and password, or use trusted connection."
        )

    if driver not in pyodbc.drivers():
        if len(pyodbc.drivers()) == 0:
            raise ValueError(
                "No ODBC drivers found. Please install an ODBC driver & specify it with the driver parameter."
            )
        else:
            raise ValueError(
                f"Driver '{driver}' not found. Pyodbc found drivers: {pyodbc.drivers()}"
            )

    # Generate the connection string:
    if trusted_cnxn:
        conn_str = (
            f"{db_type}+pyodbc://{host}/{db}?driver={driver}&trusted_connection=yes"
        )
    else:
        conn_str = (
            f"{db_type}+pyodbc://{username}:{password}@{host}/{db}?driver={driver}"
        )

    engine = create_engine(conn_str)

    return engine.connect()


def upload_data(
    df: DataFrame,
    table: str,
    cnxn: Connection,
    dtype: dict,
    override_import_dt: datetime = None,
) -> int:
    """
    Upload a DataFrame to a SQL table. Appends 'import_datetime' column to the DataFrame.

    Args:
        df (DataFrame): The DataFrame to upload.
        table (str): The name of the table to upload to.
        cnxn (Connection): The Connection object to the SQL database.
        dtype (dict): The data types of the columns in the table. Key is the column name, value is the data type as sqlalchemy.types.Something()
        override_import_dt (datetime): If provided, will override the import_datetime column with this value.

    Returns:
        int: The number of rows uploaded.
    """

    # Add an import_datetime column:
    df["import_datetime"] = (
        datetime.now() if not override_import_dt else override_import_dt
    )
    dtype["import_datetime"] = types.DateTime()

    # Upload the data:
    print(f"Uploading {len(df)} rows to {table}...")
    df.to_sql(table, cnxn, if_exists="append", index=False, dtype=dtype, chunksize=4000)

    return len(df)


def prepare_dataclass(dataclass: dataclass) -> dict:
    """
    Prepare the dataclass for insertion into a SQL database
    by converting it to a dictionary with appropriate list/dataclass
    conversions.

    Args:
        dataclass (dataclass): The dataclass to convert.

    Returns:
        dict: The dataclass converted to a dictionary.
    """

    TO_STR_FIELDS = [
        "IP_SET",
        "APPLIANCE_IDS",
        "DNS_LIST",
        "HOST_IDS",
        "NETBIOS_LIST",
        "ASSIGNED_USER_IDS",
        "ASSIGNED_UNIT_IDS",
        "EC2_IDS",
        "COMMENTS",
        "DOMAIN_LIST",
        "BUGTRAQ_LIST",
        "SOFTWARE_LIST",
        "VENDOR_REFERENCE_LIST",
        "CVE_LIST",
        "THREAT_INTELLIGENCE",
        "COMPLIANCE_LIST",
        "TAGS",
        "CLOUD_PROVIDER_TAGS",
        "IP",
        "IPV6",
        "QDS",
        "QDS_FACTORS",
        "QIDS",
        "ASSET_GROUP_TITLE",
        "ASSET_GROUP_TITLE_LIST",
        "TARGET",
        "OPTION_PROFILES",
        "REPORT_TEMPLATES",
        "REMEDIATION_POLICIES",
        "DISTRIBUTION_GROUPS",
    ]

    DICT_FIELDS = [
        "CORRELATION",
        "CVSS",
        "CVSS_V3",
        "PCI_REASONS",
        "DISCOVERY",
        "CHANGE_LOG",
        "USER_DEF",
        "TRURISK_SCORE_FACTORS",
        "VLANS",
        "ML_VERSION",
        "VULNSIGS_VERSION",
        "OPTION_PROFILE",
        "STATUS",
        "DETAILS",
        "USER",
    ]

    # Iterate over the attrs of the dataclass and convert them to the appropriate format for SQL insertion.

    for attr in dataclass.__dataclass_fields__.keys():
        if getattr(dataclass, attr):
            if attr in TO_STR_FIELDS:
                setattr(dataclass, attr, str(getattr(dataclass, attr)))
            elif attr in DICT_FIELDS and isinstance(getattr(dataclass, attr), dict):
                setattr(
                    dataclass, attr, flatten_dict_to_string(getattr(dataclass, attr))
                )

    # If there are any leftover empties or ipv4/ipv6 addresses,
    # convert them to None/str as a failsafe: #NOTE: Refactor at some point!
    for attr in dataclass.__dataclass_fields__.keys():
        if not getattr(dataclass, attr):
            setattr(dataclass, attr, None)
        elif type(getattr(dataclass, attr)) not in [str, int, float, bool, datetime]:
            setattr(dataclass, attr, str(getattr(dataclass, attr)))

    sql_dict = dataclass.to_dict()

    return sql_dict


def flatten_dict_to_string(d, parent_key="") -> str:
    """
    Format dictionary fields to a string for SQL insertion.
    """
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{k}" if parent_key else k
        if isinstance(v, dict):
            items.append(flatten_dict_to_string(v, new_key))
        else:
            items.append(f"{new_key}:{v}")
    return ", ".join(items)



"""
ips.py - IP address manipulation from Qualys subscription and stores them in a BaseList object.
"""

from typing import Union

from ..base.call_api import call_api
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.ip_converters import convert_ips, convert_ranges
from ..base.base_list import BaseList
from ..base import xml_parser


def get_ip_list(auth: BasicAuth, **kwargs) -> BaseList:
    """
    Gets a list of IP addresses from the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.

    :Kwargs:
        ```
        action (str): Action to perform on the IP addresses. Defaults to "list". WARNING: SDK automatically sets this value. It is just included for completeness.
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: SDK automatically sets this value. It is just included for completeness.
        ips (str): Show only certain IP addresses/ranges. One or more IPs/ranges may be specified. Multiple entries are comma separated. A host IP range is specified with a hyphen (for example, 10.10.10.44-10.10.10.90).
        network_id (Union[str, int]): Network ID to filter on. Defaults to None. NOTE: This has to be enabled in the Qualys subscription!
        tracking_method (Literal[None, "IP", "DNS", "NETBIOS"]): Tracking method to filter on. Defaults to None. Valid values are IP, DNS, NETBIOS.
        compliance_enabled (Union[str, bool]): Whether to filter to IPs within the compliance module. Defaults to None.
        certview_enabled (Union[str, bool]): Whether to filter to IPs within certview module. Defaults to None.

    Returns:
        BaseList: BaseList object containing the IP addresses/ranges.
    """
    ip_list = BaseList()

    kwargs["action"] = "list"
    kwargs["echo_request"] = False

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="get_ip_list",
        params=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    if response.status_code == 200:
        data = xml_parser(response.text)["IP_LIST_OUTPUT"]

        if "IP_SET" not in data["RESPONSE"]:
            print("No IP addresses found. Returning empty BaseList.")
            return ip_list

        data = data["RESPONSE"][
            "IP_SET"
        ]  # at this point, data has IP and IP_RANGE keys

        # Convert the IP addresses into IP objects:
        if "IP" in data:
            # Normalize, account for a str:
            if isinstance(data["IP"], str):
                data["IP"] = [data["IP"]]
            single_ips = [ip for ip in data["IP"]]  # Single IPs
            ip_list.extend(convert_ips(single_ips))

        # Convert the IP ranges into IPNetwork objects:
        if "IP_RANGE" in data:
            # Normalize, account for a str:
            if isinstance(data["IP_RANGE"], str):
                data["IP_RANGE"] = [data["IP_RANGE"]]
            range_ips = [ip for ip in data["IP_RANGE"]]  # IP Ranges
            ip_list.extend(convert_ranges(range_ips))

    else:
        raise Exception(f"Failed to pull IP list. Status code: {response.status_code}")

    return ip_list


def add_ips(
    auth: BasicAuth,
    ips: Union[str, BaseList],
    enable_pc: Union[bool, int] = False,
    enable_vm: Union[bool, int] = True,
    enable_sca: Union[bool, int] = False,
    **kwargs,
) -> None:
    """
    Adds IP addresses to the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        ips (Union[str, BaseList): List of IP addresses/ranges to add. Can be a string or a BaseList object containing IPs. Multiple entries are comma separated. A host IP range is specified with a hyphen.
        enable_pc (Union[bool, int]): Whether to enable policy compliance tracking on the IP addresses. Defaults to False.
        enable_vm (Union[bool, int]): Whether to enable vulnerability management tracking on the IP addresses. Defaults to True.
        enable_sca (bool): Whether to enable SCA on the IP addresses. Defaults to False.
        NOTE: EITHER enable_pc OR enable_vm MUST BE TRUE FOR THE IP ADDITION TO WORK!

    :Kwargs:
        ```
        action (str): Action to perform on the IP addresses. Defaults to "add". WARNING: SDK automatically sets this value. It is just included for completeness.
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: SDK automatically sets this value. It is just included for completeness.
        tracking_method (Literal["IP", "DNS", "NETBIOS"]): Tracking method to filter on. Defaults to "IP". Valid values are IP, DNS, NETBIOS.
        owner (str): Owner of the IP addresses. Defaults to None.
        ud1 (str): User-defined field 1. Defaults to None.
        ud2 (str): User-defined field 2. Defaults to None.
        ud3 (str): User-defined field 3. Defaults to None.
        comment (str): Comment for the IP addresses. Defaults to None.
        ag_title (str): Asset group title to add the IP addresses to. Defaults to None. Required if user is a Unit Manager.
        enable_certview (bool): Whether to enable CertView on the IP addresses. Defaults to False.
        ```

    Returns:
        None
    """

    # Check if either enable_pc or enable_vm is True:
    if not enable_pc and not enable_vm:
        raise ValueError("Either enable_pc or enable_vm must be True!")

    # Check tracking_method:
    if "tracking_method" in kwargs and kwargs["tracking_method"] not in [
        "IP",
        "DNS",
        "NETBIOS",
    ]:
        raise ValueError(
            f"Invalid tracking method. Valid values are IP, DNS, NETBIOS, not {kwargs['tracking_method']}."
        )

    kwargs["action"] = "add"
    kwargs["echo_request"] = False

    # If ips is a BaseList object, convert it to a list of strings:
    if isinstance(ips, BaseList):
        ips = [str(ip) for ip in ips]
        ips = ",".join(ips)

    kwargs["ips"] = ips
    kwargs["enable_pc"] = enable_pc
    kwargs["enable_vm"] = enable_vm
    kwargs["enable_sca"] = enable_sca

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="add_ips",
        payload=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    result = xml_parser(response.text)["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]

    print(result)


def update_ips(auth: BasicAuth, ips: Union[str, BaseList], **kwargs) -> None:
    """
    Update specific details of IP addresses in the Qualys subscription.

    Params:
        auth (BasicAuth): Qualys BasicAuth object.
        ips (Union[str, BaseList): List of IP addresses/ranges to update. Can be a string or a BaseList object containing IPs. Multiple entries are comma separated. A host IP range is specified with a hyphen.

    :Kwargs:
        ```
        action (str): Action to perform on the IP addresses. Defaults to "update". WARNING: SDK automatically sets this value. It is just included for completeness.
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: SDK automatically sets this value. It is just included for completeness.
        network_id (Union[str, int]): Network ID to filter on. Defaults to None. NOTE: This has to be enabled in the Qualys subscription!
        tracking_method (Literal["IP", "DNS", "NETBIOS"]): Tracking method to filter on. Defaults to "IP". Valid values are IP, DNS, NETBIOS.
        host_dns (str): The DNS hostname for the IP you want to update. A single IP must be specified in the same request and the IP will only be updated if it matches the hostname specified.
        host_netbios (str): The NetBIOS hostname for the IP you want to update. A single IP must be specified in the same request and the IP will only be updated if it matches the hostname specified.
        owner (str): Owner of the IP addresses. Defaults to None.
        ud1 (str): User-defined field 1. Defaults to None.
        ud2 (str): User-defined field 2. Defaults to None.
        ud3 (str): User-defined field 3. Defaults to None.
        comment (str): Comment for the IP addresses. Defaults to None.

    Returns:
        None
    """

    kwargs["action"] = "update"
    kwargs["echo_request"] = False

    # If ips is a BaseList object, convert it to a list of strings:
    if isinstance(ips, BaseList):
        ips = [str(ip) for ip in ips]
        ips = ",".join(ips)

    kwargs["ips"] = ips

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="update_ips",
        payload=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    result = xml_parser(response.text)["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]

    print(result)
